#!/usr/bin/env python

"""
@copyright: Copyright (C) 2008, 2013 Oracle.  All rights reserved.

@license:   This program is free software; you can redistribute it and/or
            modify it under the terms of the GNU General Public
            License as published by the Free Software Foundation, version 2.

            This program is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY; without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
            General Public License for more details.

            You should have received a copy of the GNU General Public
            License along with this program; if not, write to the Free Software
            Foundation, Inc., 59 Temple Place - Suite 330, Boston,
            MA 021110-1307, USA

@summary:   This is the implementation of the Oracle VM Storage Connect
            Plug-in for Generic iSCSI and FibreChannel.
"""

__docformat__ = 'epytext en'

import gettext
import subprocess
import platform

from OSCPlugin import (ABILITY_TYPES, IPlugin, IStorageArrayPlugin,
                       SE_STATES, SE_STATES_TEXT,
                       InvalidValueEx, MissingKeyEx, NoSuchOperationEx,
                       OperationFailedEx)

#
# Import the module that provides the appropriate OS specific functions
# based on the operating system we are running on.
#

MODULES = {
    'Linux': { 'scsi': 'scsi_linux', 'san': 'san_linux' },
    'SunOS': { 'scsi': 'scsi_solaris', 'san': 'san_solaris' }
}
            

scsi = __import__(MODULES[platform.system()]['scsi'], globals(), locals(), [])
san = __import__(MODULES[platform.system()]['san'], globals(), locals(), [])

SCSI_GOOD = 0
SCSI_CHECK = 5

vg_GenericISCSI = "Generic_iSCSI_Volume_Group"
vg_GenericSAN = "Generic_SAN_Volume_Group"

#
# Private module variables
#
_ = gettext.gettext
__min_osc_version__ = ["1", "2", "7"]

class GenericPlugin(IStorageArrayPlugin):
    """
    This is the Oracle Generic FC / iSCSI Generic Plugin
    """

    #
    # Required class variables as per the Storage Connect Plugin spec
    #
    plugin_name             = "Oracle Generic SCSI Plugin"
    vendor_name             = "Oracle"
    plugin_version          = "1.1.0"
    plugin_desc             = "Oracle Storage Connect Plugin for Generic FC and iSCSI"
    required_api_vers       = __min_osc_version__
    extended_api_version    = ["1", "2", "9"]
    storage_types           = [IStorageArrayPlugin.SANStorage,
                               IStorageArrayPlugin.iSCSIStorage]
    plugin_ability          = {"snapshot":              ABILITY_TYPES.UNSUPPORTED,
                               "custom_snap_name":      ABILITY_TYPES.UNSUPPORTED,
                               "snap_is_sync":          ABILITY_TYPES.UNSUPPORTED,
                               "clone":                 ABILITY_TYPES.UNSUPPORTED,
                               "custom_clone_name":     ABILITY_TYPES.UNSUPPORTED,
                               "clone_is_sync":         ABILITY_TYPES.UNSUPPORTED,
                               "resize":                ABILITY_TYPES.UNSUPPORTED,
                               "resize_is_sync":        ABILITY_TYPES.UNSUPPORTED,
                               "splitclone":            ABILITY_TYPES.UNSUPPORTED,
                               "splitclone_is_sync":    ABILITY_TYPES.UNSUPPORTED,
                               "snapclone":             ABILITY_TYPES.UNSUPPORTED,
                               "snapclone_is_sync":     ABILITY_TYPES.UNSUPPORTED,
                               "require_storage_name":  ABILITY_TYPES.UNSUPPORTED,
                               "access_control":        ABILITY_TYPES.NO}

    __generic_plugin__      = True

    @staticmethod
    def validate(ss_record):
        IPlugin.logger.debug("ENTER: validate()")
        missing_key = ""

        if not ss_record.get("uuid", None):
            missing_key += " uuid"

        if not ss_record.get("name", None):
            missing_key += " name"

        if missing_key:
            IPlugin.logger.error("validate() - '%s' have to be supplied!" %
                                 missing_key)
            raise MissingKeyEx(_("'%s' have to be supplied" % missing_key),
                                 missing_key)

        if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage: 
            IPlugin.logger.debug("IStorageArrayPlugin.iSCSIStorage")

        elif ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:
            IPlugin.logger.debug("IStorageArrayPlugin.SANStorage")

        else:
            IPlugin.logger.error("validate() - Unknown storage_type %s "
                                 "specified." % ss_record["storage_type"])
            raise InvalidValueEx(_("Unknown storage_type %s specified" %
                                   ss_record["storage_type"]))


        IPlugin.logger.debug("EXIT : validate() - Success")


    @staticmethod
    def getCapabilities(ss_record):
        IPlugin.logger.debug("ENTER: getCapabilities()")
        IPlugin.logger.debug("EXIT : getCapabilities()")
        return GenericPlugin.plugin_ability


    @staticmethod
    def getInfo(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: getInfo()")

        if se_record:

            if not se_record.get("uuid", None):
                IPlugin.logger.error("getInfo - No uuid")
                raise InvalidValueEx(_("No uuid"))


            try:
                lun_info = scsi.scsi_info(se_record["page83_id"])

            except IOError, e:
                IPlugin.logger.error("IOError: Exception %s obtaining LUN "
                                     "info for %s",
                                     e.strerror, se_record["page83_id"])
                raise OperationFailedEx(_("IOError: Exception %s obtaining "
                                          "LUN info for %s") %
                                        (e.strerror, se_record["page83_id"]))

            se_record["se_type"] = IStorageArrayPlugin.LUNType


            se_record["id"] = lun_info["serial"]

            se_record["page83_id"] = lun_info["page83_id"]

            se_record["size"] = lun_info["size"]

            se_record["status"] = GenericPlugin.getStatus(ss_record, se_record)
            
            se_record["state"] = GenericPlugin.getState(ss_record, se_record)

            se_record["path"] = [lun_info["dev_name"]]

            if se_record.get("vol_group_name", None):
                ss_volgroups = ss_record.get('vol_groups', []) 
                if len(ss_volgroups) > 0:
                    se_record["vol_group_name"] = ss_volgroups[0] 
                else:
                    if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                        se_record["vol_group_name"] = vg_GenericISCSI
                    else:
                        se_record["vol_group_name"] = vg_GenericSAN

            IPlugin.logger.debug("EXIT : getInfo() - Success")
            return se_record

        else:
            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                ss_record["storage_desc"] = "Generic iSCSI Storage"

            elif ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:
                ss_record["storage_desc"] = "Generic Fibre Channel Storage"

            ss_record["status"] = GenericPlugin.getStatus(ss_record)

            groups = []

            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                groups.append({
                              "vol_name":      vg_GenericISCSI,
                              "vol_total_sz":  0,
                              "vol_used_sz":   0,
                              "vol_free_sz":   0,
                              "vol_alloc_sz":  0,
                              })
            else:
                groups.append({
                              "vol_name":      vg_GenericSAN,
                              "vol_total_sz":  0,
                              "vol_used_sz":   0,
                              "vol_free_sz":   0,
                              "vol_alloc_sz":  0,
                              })

            ss_record["vol_groups"] = groups

            IPlugin.logger.debug("EXIT : getInfo() - Success")
            return ss_record


    @staticmethod
    def list(ss_record, se_type = IPlugin.LUNType):

        IPlugin.logger.debug("ENTER: list()")
        se_record_list = []

        luns = scsi.scsi_luns(ss_record)

        if not luns:
            IPlugin.logger.debug("EXIT : list() - Nothing to do!")
            return se_record_list

        if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
            vg = vg_GenericISCSI
        else:
            vg = vg_GenericSAN

        for lun_info in luns:

            se_type = IStorageArrayPlugin.LUNType
            se_record = {"se_type":         se_type,
                         "ss_uuid":         ss_record["uuid"],
                         "uuid":            lun_info["page83_id"],
                         "id":              lun_info["serial"],
                         "path":            [san.device_path(lun_info["page83_id"])],
                         "size":            lun_info["size"],
                         "page83_id":       lun_info["page83_id"],
                         "vol_group_name":  vg,}

            se_record["status"] = GenericPlugin.getStatus(ss_record, se_record)
            
            se_record["state"] = GenericPlugin.getState(ss_record, se_record)

            se_record_list.append(se_record)

        IPlugin.logger.debug("EXIT : list() - Success")
        return se_record_list


    @staticmethod
    def updateSERecords(ss_record, se_records, append_allowed = False):
        IPlugin.logger.debug("ENTER: updateSERecord()")

        if len(se_records) < 1 :
            IPlugin.logger.debug("EXIT : updateSERecord() - Nothing to do!")
            return se_records

        for se_record in se_records:

            if dev_name == 'control':
                continue

            try:
                se_record = GenericPlugin.getInfo(ss_record, se_record)

            except IOError, e:
                IPlugin.logger.error("IOError: Exception %s obtaining LUN "
                                     "info from %s/%s",
                                     e.strerror,
                                     se_record["uuid"])
                continue


        IPlugin.logger.debug("EXIT : updateSERecord() - Success")
        return se_records


    @staticmethod
    def getStatus(ss_record, se_record = None):

        #IPlugin.logger.debug("ENTER: getStatus()")

        if not se_record:
            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
                status = scsi.iscsi_service_status()

            elif ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:
                status = GenericPlugin._get_fc_status()

            else:
                raise InvalidValueEx(_("Unsupported storage type is specified"))
        else:
            if se_record["se_type"] == IStorageArrayPlugin.LUNType:
                status = scsi.scsi_status(se_record["uuid"])
                if not status:
                    status = "no status"
            else:
                raise InvalidValueEx(_("Unsupported storage element type is specified"))

        #IPlugin.logger.debug("EXIT : getStatus() - Success")

        return status

    @staticmethod
    def getState(ss_record, se_record):

        #IPlugin.logger.debug("ENTER: getState()")

        if se_record["se_type"] == IStorageArrayPlugin.LUNType:
            state = scsi.scsi_state(se_record["uuid"])
            if not state:
                state = SE_STATES.UNKNOWN
        else:
            raise InvalidValueEx(_("Unsupported storage element type is specified"))

        #IPlugin.logger.debug("EXIT : getState() - Success")

        return state

    @staticmethod
    def getCloneLimits(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: getCloneLimits()")
        IPlugin.logger.debug("EXIT : getCloneLimits()")
        return -1

    @staticmethod
    def isCloneable(ss_record, se_record):
        IPlugin.logger.debug("ENTER: isCloneable()")
        if se_record["se_type"] == IStorageArrayPlugin.LUNType:
            IPlugin.logger.debug("EXIT : isCloneable() - True")
            return True

        IPlugin.logger.debug("EXIT : isCloneable() - False")
        return False


    @staticmethod
    def copy(ss_record, se_record, dest_se_record, qos = None):
        IPlugin.logger.debug("ENTER: copy()")

        p_cmd = ("/bin/dd "
               "if=%s "
               "of=%s "
               "bs=4096k "
               "oflag=direct") % (se_record["path"], dest_se_record["path"])

        p_cmd = p_cmd.split(" ")
        dd_p = subprocess.Popen(p_cmd,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE,
                                close_fds=True)
        (p_out, p_err) = dd_p.communicate()

        if dd_p.returncode != 0:
            raise IOError("Unable to copy %s: %s" % (se_record["path"], p_err))

        dest_se_record = GenericPlugin.getInfo(ss_record, dest_se_record)

        IPlugin.logger.debug("EXIT : copy() - Success")
        return dest_se_record

    @staticmethod
    def getCurrentClones(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: getCurretClones()")
        IPlugin.logger.debug("EXIT :getCurrentClones() - Success")
        return []

    @staticmethod
    def addToAccessGroup(ss_record, access_grp_name, grp_entries):
        IPlugin.logger.debug("ENTER: addToAccessGroup()")
        IPlugin.logger.error("addToAccessGroup() is not supported!")
        raise NoSuchOperationEx(_("addToAccessGroup() is not supported!"))


    @staticmethod
    def removeFromAccessGroup(ss_record, access_grp_name, grp_entries):
        IPlugin.logger.debug("ENTER: removeAccessGroups()")
        IPlugin.logger.error("removeAccessGroups() is not supported!")
        raise NoSuchOperationEx(_("removeAccessGroups() is not supported!"))

    @staticmethod
    def getAccessGroups(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: getAccessGroups()")
        IPlugin.logger.error("getAccessGroupgs() is not supported!")
        raise NoSuchOperationEx(_("getAccessGroups() is not supported!"))


    @staticmethod
    def createAccessGroups(ss_record, access_grps):
        IPlugin.logger.debug("ENTER: createAccessGroups()")
        IPlugin.logger.error("createAccessGroups() is not supported!")
        raise NoSuchOperationEx(_("createAccessGroups() is not supported!"))


    @staticmethod
    def removeAccessGroups(ss_record, access_grps):
        IPlugin.logger.debug("ENTER: removeAccessGroups()")
        IPlugin.logger.error("removeAccessGroups() is not supported!")
        raise NoSuchOperationEx(_("removeAccessGroups() is not supported!"))

    @staticmethod
    def getStorageNames(ss_record):
        IPlugin.logger.debug("ENTER: getStorageNames()")
        IPlugin.logger.error("getStorageNames() is not supported!")
        raise NoSuchOperationEx(_("getStorageNames() is not supported!"))

    @staticmethod
    def stopPresent(ss_record, se_record):
        IPlugin.logger.debug("ENTER: stopPresent()")
        IPlugin.logger.error("stopPresent() is not supported!")
        raise NoSuchOperationEx(_("stopPresent() is not supported!"))

    @staticmethod
    def online(ss_record, se_record):
         IPlugin.logger.debug("ENTER: online()")
         IPlugin.logger.error("online() is not supported!")
         raise NoSuchOperationEx(_("online() is not supported!"))

    @staticmethod
    def offline(ss_record, se_record):
        IPlugin.logger.debug("ENTER: offline()")
        IPlugin.logger.error("offline() is not supported!")
        raise NoSuchOperationEx(_("offline() is not supported!"))


    @staticmethod
    def create(ss_record, se_record, qos = None):
        IPlugin.logger.debug("ENTER: create()")
        IPlugin.logger.error("create() is not supported!")
        raise NoSuchOperationEx(_("create() is not supported!"))


    @staticmethod
    def startPresent(ss_record, se_record):
        IPlugin.logger.debug("ENTER: startPresent()")
        IPlugin.logger.error("startPresent() is not supported!")
        raise NoSuchOperationEx(_("startPresent() is not supported!"))


    @staticmethod
    def resize(ss_record, se_record):
        IPlugin.logger.debug("ENTER: resize()")
        IPlugin.logger.error("resize() is not supported!")
        raise NoSuchOperationEx(_("resize() is not supported!"))


    @staticmethod
    def remove(ss_record, se_record):
        IPlugin.logger.debug("ENTER: remove()")
        IPlugin.logger.error("remove() is not supported!")
        raise NoSuchOperationEx(_("remove() is not supported!"))

    @staticmethod
    def splitClone(ss_record, se_record):
        IPlugin.logger.debug("ENTER: splitClone()")
        IPlugin.logger.error("splitClone() is not supported!")
        raise NoSuchOperationEx(_("splitClone() is not supported!"))


    @staticmethod
    def cloneFromSnap(ss_record, se_record, snap_se_record, dest_se_record = None, qos = None):
        IPlugin.logger.debug("ENTER: cloneFromSnap()")
        IPlugin.logger.error("cloneFromSnap() is not supported!")
        raise NoSuchOperationEx(_("(cloneFromSnap) is not supported!"))


    @staticmethod
    def getSnapLimits(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: getSnapLimits()")
        IPlugin.logger.error("getSnapLimits() is not supported!")
        raise NoSuchOperationEx(_("getSnapLimits() is not supported!"))


    @staticmethod
    def isSnapable(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: isSnapable()")
        IPlugin.logger.error("isSnapable() is not supported!")
        raise NoSuchOperationEx(_("isSnapable() is not supported!"))


    @staticmethod
    def createSnap(ss_record, se_record, snap_se_record = None):
        IPlugin.logger.debug("ENTER: createSnap()")
        IPlugin.logger.error("createSnap() is not supported!")
        raise NoSuchOperationEx(_("createSnap() is not supported!"))


    @staticmethod
    def createMultiSnap(ss_record, se_records, snap_se_record = None):
        IPlugin.logger.debug("ENTER: createMultiSnap()")
        IPlugin.logger.error("(createMultiSnap() is not supported!")
        raise NoSuchOperationEx(_("createMultiSnap(() is not supported!"))


    @staticmethod
    def isRestorable(ss_record, snap_se_record, se_record = None):
        IPlugin.logger.error("isRestorable() is not supported!")
        raise NoSuchOperationEx(_("isRestorable() is not supported!"))


    @staticmethod
    def snapRestore(ss_record, snap_se_record, se_record = None):
        IPlugin.logger.debug("ENTER: snapRestore()")
        IPlugin.logger.error("snapRestore() is not supported!")
        raise NoSuchOperationEx(_("snapRestore() is not supported!"))


    @staticmethod
    def snapRemove(ss_record, snap_se_records):
        IPlugin.logger.debug("ENTER: snapRemove()")
        IPlugin.logger.error("snapRemove() is not supported!")
        raise NoSuchOperationEx(_("snapRemove() is not supported!"))


    @staticmethod
    def getCurrentSnaps(ss_record, se_record = None):
        IPlugin.logger.debug("ENTER: getCurrentSnaps()")
        IPlugin.logger.error("getCurrentSnaps() is not supported!")
        raise NoSuchOperationEx(_("getCurrentSnaps() is not supported!"))


    @staticmethod
    def getQoSList(ss_record):
        IPlugin.logger.debug("ENTER: getQoSList()")
        IPlugin.logger.error("getQoSList() is not supported!")
        raise NoSuchOperationEx(_("getQoSList() is not supported!"))


    @staticmethod
    def setQoS(ss_record, se_record, qos):
        IPlugin.logger.debug("ENTER: setQoS()")
        IPlugin.logger.error("setQoS() is not supported!")
        raise NoSuchOperationEx(_("setQoS() is not supported!"))



    #########################################################################
    ###             START OF THE PLUGIN PRIVATE METHODS                   ###
    #########################################################################


    @staticmethod
    def _get_fc_status():
        IPlugin.logger.debug("ENTER: _get_fc_status()")

        return "online"

    @staticmethod
    def _get_fc_state():
        IPlugin.logger.debug("ENTER: _get_fc_state()")

        return SE_STATES.ONLINE

