#!/usr/bin/env python

"""
@copyright: Copyright (C) 2013 Oracle.  All rights reserved.

@license:   This program is free software; you can redistribute it and/or
            modify it under the terms of the GNU General Public
            License as published by the Free Software Foundation, version 2.

            This program is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY; without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
            General Public License for more details.

            You should have received a copy of the GNU General Public
            License along with this program; if not, write to the Free Software
            Foundation, Inc., 59 Temple Place - Suite 330, Boston,
            MA 021110-1307, USA

@summary:   This provides Linux specific functions for the Oracle VM
            Storage Connect Plug-in for Generic FibreChannel.

"""
import gettext
import subprocess
import os

import OSCPluginUtils

from cStringIO import StringIO
from OSCPlugin import (IPlugin, IStorageArrayPlugin,
                       SE_STATES, SE_STATES_TEXT,
                       OperationFailedEx, InvalidValueEx)


_ = gettext.gettext


def _get_lun_info_(dev_name):

    lun_info = {"dev_name": dev_name}

    # This is the SG_IO INQUIRY CDB
    get_sg_inq = "12 00 00 00 60 00"
    p_cmd = ("/usr/bin/sg_raw "
             "--binary "
             "--nosense "
             "--request=96 "
             "--timeout=15 "
             "%s " +
             get_sg_inq) % dev_name

    p_cmd = p_cmd.split(" ")
    sgraw_p = subprocess.Popen(p_cmd,
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                               close_fds=True)
    (p_out, p_err) = sgraw_p.communicate()

    if sgraw_p.returncode != 0:
        IPlugin.logger.debug("Unable to issue SG_IO INQ CDB: %s" % p_err)
        raise IOError("Unable to issue SG_IO INQ CDB: %s" % p_err)

    # Use StringIO to read from the string
    str_rdr = StringIO(p_out)
    str_rdr.seek(8)

    vendor  = str_rdr.read(8)
    product = str_rdr.read(16)
    version = str_rdr.read(4)
    serial  = str_rdr.read(32)

    lun_info["vendor"]  = vendor.strip()
    lun_info["product"] = product.strip()
    lun_info["version"] = version.strip()


   # This is the SG_IO INQUIRY CDB Page 80
    get_sg_inq = "12 01 80 00 fc 00"
    p_cmd = ("/usr/bin/sg_raw "
             "--binary "
             "--nosense "
             "--request=256 "
             "--timeout=15 "
             "%s " +
             get_sg_inq) % dev_name

    p_cmd = p_cmd.split(" ")
    sgraw_p = subprocess.Popen(p_cmd,
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                               close_fds=True)
    (p_out, p_err) = sgraw_p.communicate()

    if sgraw_p.returncode != 0:
        IPlugin.logger.debug("Unable to issue SG_IO INQ Page 80 CDB to %s: %s" % (dev_name, p_err))
        raise IOError("Unable to issue SG_IO INQ Page 80 CDB to %s: %s" % (dev_name, p_err))

    # Use StringIO to read from the string
    str_rdr = StringIO(p_out)
    str_rdr.seek(4)

    serial  = str_rdr.read(32)

    lun_info["serial"]  = serial[:serial.find('\0')].strip()


    # Use blockdev to get size from the device
    p_cmd = ("/sbin/blockdev "
             "--getsz "
             "%s") % dev_name

    p_cmd = p_cmd.split(" ")
    blkdev_p = subprocess.Popen(p_cmd,
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                               close_fds=True)
    (p_out, p_err) = blkdev_p.communicate()

    if blkdev_p.returncode != 0:
        IPlugin.logger.debug("Unable to retrieve blockdev info: %s" % p_err)
        raise IOError("Unable to retrieve blockdev info: %s" % p_err)

    size = 0
    size = int(p_out)
    lun_info["size"] = size * 512

    if os.path.exists("/dev/emcpower"):

        # Get page 0x83: VPD info 12 01 83 00 fc 00
        get_sg_inq = "12 01 83 00 fc 00"
        p_cmd = ("/usr/bin/sg_raw "
                 "--binary "
                 "--nosense "
                 "--request=256 "
                 "--timeout=15 "
                 "%s " +
                 get_sg_inq) % dev_name


        p_cmd = p_cmd.split(" ")
        sgraw_p = subprocess.Popen(p_cmd,
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)

        (p_out, p_err) = sgraw_p.communicate()

        if sgraw_p.returncode != 0:
            IPlugin.logger.debug("Unable to issue SG_IO INQ Page 83 CDB to %s: %s" % (dev_name, p_err))
            raise IOError("Unable to issue SG_IO INQ Page 83 CDB to %s: %s" % (dev_name, p_err))

        page_83 = []
        for index in range(len(p_out)):
            page_83.append(ord(p_out[index]))

        lun_info["page83_id"] = OSCPluginUtils.makeMPPage83FromRawPage83(page_83)

    else:

        lun_info["page83_id"] = OSCPluginUtils.getPage83FromMPDev(dev_name)


    # deactivated code for now
    if 0:

        # get paths to the device
        p_cmd = ("/bin/ls "
                 "-l "
                 "%s") % dev_name

        p_cmd = p_cmd.split(" ")

        path_p = subprocess.Popen(p_cmd,
                                  stdout = subprocess.PIPE,
                                  stderr = subprocess.PIPE,
                                  close_fds=True)
        (p_out, p_err) = path_p.communicate()

        if path_p.returncode != 0:
            raise IOError("Unable to list paths for device %s: %s" % (dev_name, p_err))

        index = p_out.find('->')
        if index < 0:
            return lun_info

        parent_dev = p_out[index+6:-1]

        lun_info["path"] = []
        lun_info["path"].append("/dev/%s" % parent_dev)

        for device in os.listdir("/sys/block/%s/slaves/" % parent_dev):
            lun_info["path"].append("/dev/%s" % device)

    # end of deactivate

    return lun_info


def _get_dev_server(ss_record, mp_dev):

    disk = None

    dm = os.path.realpath("%s/%s" % (IPlugin.dev_path_prefix, mp_dev))

    dm = dm.replace('/dev/', '')

    f = open("/sys/block/" + dm + "/dm/uuid")
    uuid = f.read().strip()
    f.close()
    if not uuid.startswith("mpath-"):
        return False

    for dev in os.listdir("/sys/block/" + dm + "/slaves/"):
        if dev.startswith('sd'):
            disk =  os.listdir("/sys/block/" + dm + "/slaves/" + dev + "/device/scsi_disk")
            break

    if disk:  # We only need one slave to get the info
        ls_p = subprocess.Popen(["lsscsi",
                                  "-t",
                                  disk[0]],
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE,
                                 close_fds=True)
        (p_out, p_err) = ls_p.communicate()

        if ls_p.returncode != 0:
            IPlugin.logger.error("Unable to list device %s: %s" %
                                 (mp_dev, p_err))
            raise OperationFailedEx(_("Unable to list device %s: %s" % (mp_dev, p_err)))

        ls_fields = p_out.split()

        if ls_fields[2].startswith('iqn.'):
            if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:

                host_target = ls_fields[2].split(',')

                if ss_record.get("iscsi_storage_targets", None):
                    host_target = ls_fields[2].split(',')
                    for target in ss_record["iscsi_storage_targets"]:
                        if target['iscsi_target_name'] == host_target[0]:
                            return True

        else:
            if ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:
                return True

    return False


def scsi_luns(ss_record):

    if not os.path.exists(IPlugin.dev_path_prefix):
        return []

    luns = []

    for dev_name in os.listdir(IPlugin.dev_path_prefix):

        if dev_name == 'control':
            continue

        try:
            lun_info = _get_lun_info_("%s/%s" % (IPlugin.dev_path_prefix, dev_name))

        except IOError, e:
            IPlugin.logger.error("IOError: Exception %s obtaining LUN "
                                 "info from %s/%s",
                                 e.strerror,
                                 IPlugin.dev_path_prefix,
                                 dev_name)
            continue

        if not os.path.exists("/dev/emcpower"):
            if not _get_dev_server(ss_record, dev_name):
                continue

        luns.append(lun_info)

    return luns


def scsi_info(disk_id):

    dev_name = "%s/%s" % (IPlugin.dev_path_prefix, disk_id)

    if not os.path.exists(dev_name):
        IPlugin.logger.error("Device %s does not exist", dev_name)
        raise InvalidValueEx(_("Device %s does not exist" % dev_name))

    return _get_lun_info_(dev_name)


def scsi_status(disk_id):

    status = None

    dev_name = "%s/%s" % (IPlugin.dev_path_prefix, disk_id)

    # This is the SG_IO TUR CDB
    get_sg_tur = "00 00 00 00 00 00"
    p_cmd = ("/usr/bin/sg_raw "
             "--binary "
             "--nosense "
             "--timeout=15 "
             "%s " +
             get_sg_tur) % dev_name
    p_cmd = p_cmd.split(" ")
    sgraw_p = subprocess.Popen(p_cmd,
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                               close_fds=True)

    (p_out, p_err) = sgraw_p.communicate()

    if p_err.find('Good') >= 0:
       # IPlugin.logger.debug("EXIT : getStatus() - Success")
        status = 'online'

    if p_err.find('Check') >= 0:
       # IPlugin.logger.debug("EXIT : getStatus() - Success")
        status = 'offline'

    if sgraw_p.returncode != 0:
        raise IOError("Unable to issue SG_IO TUR CDB to %s: %s" % (dev_name, p_err))

    return status


def scsi_state(disk_id):

    # No good way to get the se_record.state for a generic plug-in 
    # so just use the status.
    if scsi_status(disk_id) == "online":
        state = SE_STATES.ONLINE
    else:
        state = SE_STATES.OFFLINE

    return state


def iscsi_service_status():
    iscsi_p = subprocess.Popen(["service",
                                "iscsi",
                                "status"],
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                                close_fds=True)
    (p_out, p_err) = iscsi_p.communicate()

    if p_out.find('running') < 0:
        return "offline"

    return "online"
